// import bootstrap from 'https://cdn.jsdelivr.net/npm/bootstrap@5.3.3/+esm';
import markJs from 'https://cdn.jsdelivr.net/npm/mark.js@8.11.1/+esm';
import TomSelect from 'https://cdn.jsdelivr.net/npm/tom-select@2.3.1/+esm';
import basicLightbox from 'https://cdn.jsdelivr.net/npm/basiclightbox@5.0.4/+esm';
import slugify from 'https://cdn.jsdelivr.net/npm/slugify@1.6.6/+esm';
import DirtyForm from 'https://cdn.jsdelivr.net/npm/dirty-form@0.4.0/+esm';
import Trix from 'https://cdn.jsdelivr.net/npm/trix@2.1.0/+esm';

// Provide Bootstrap variable globally to allow custom backend pages to use it
// window.bootstrap = bootstrap;

document.addEventListener('DOMContentLoaded', () => {
    window.EasyAdminApp = new App();
    console.log('Class App initialized');

    // toggle switches are only created in index page (i.e. in datagrid tables) because
    // in other pages they act like simple checkboxes or labels. Only in index page
    // the toggle switch can change the value of an entity propert via Ajax requests
    document.querySelectorAll('td.field-boolean .form-switch input[type="checkbox"]').forEach((toggleField) => {
        new ToggleSwitch(toggleField);
    });

    document.querySelectorAll('[data-ea-slug-field]').forEach((field) => {
        new Slugger(field);
    });

    document.querySelectorAll('[data-ea-textarea-field]').forEach(function (field) {
        new Autogrow(field);
    });

    document.querySelectorAll('.ea-lightbox-thumbnail').forEach((imageElement) => {
        new Image(imageElement);
    });

    new Form();
});

/* -------------------------------------------------------------------------- */
/*                             FIELD-COLLECTION.JS                            */
/* -------------------------------------------------------------------------- */
const eaCollectionHandler = function (event) {
    document.querySelectorAll('button.field-collection-add-button').forEach((addButton) => {
        const collection = addButton.closest('[data-ea-collection-field]');

        if (!collection || collection.classList.contains('processed')) {
            return;
        }

        EaCollectionProperty.handleAddButton(addButton, collection);
        EaCollectionProperty.updateCollectionItemCssClasses(collection);
    });

    document.querySelectorAll('button.field-collection-delete-button').forEach((deleteButton) => {
        deleteButton.addEventListener('click', () => {
            const collection = deleteButton.closest('[data-ea-collection-field]');
            const item = deleteButton.closest('.field-collection-item');

            item.remove();
            document.dispatchEvent(new Event('ea.collection.item-removed'));

            EaCollectionProperty.updateCollectionItemCssClasses(collection);
        });
    });
}

window.addEventListener('DOMContentLoaded', eaCollectionHandler);
document.addEventListener('ea.collection.item-added', eaCollectionHandler);

const EaCollectionProperty = {
    handleAddButton: (addButton, collection) => {
        addButton.addEventListener('click', function() {
            const isArrayCollection = collection.classList.contains('field-array');
            // Use a counter to avoid having the same index more than once
            let numItems = parseInt(collection.dataset.numItems);

            // Remove the 'Empty Collection' badge, if present
            const emptyCollectionBadge = this.parentElement.querySelector('.collection-empty');
            if (null !== emptyCollectionBadge) {
                emptyCollectionBadge.outerHTML = isArrayCollection ? '<div class="ea-form-collection-items"></div>' : '<div class="ea-form-collection-items"><div class="accordion"><div class="form-widget-compound"><div data-empty-collection></div></div></div></div>';
            }

            const formTypeNamePlaceholder = collection.dataset.formTypeNamePlaceholder;
            const labelRegexp = new RegExp(formTypeNamePlaceholder + 'label__', 'g');
            const nameRegexp = new RegExp(formTypeNamePlaceholder, 'g');

            let newItemHtml = collection.dataset.prototype
                .replace(labelRegexp, numItems)
                .replace(nameRegexp, numItems);

            collection.dataset.numItems = ++numItems;
            const newItemInsertionSelector = isArrayCollection ? '.ea-form-collection-items' : '.ea-form-collection-items .accordion > .form-widget-compound [data-empty-collection]';
            const collectionItemsWrapper = collection.querySelector(newItemInsertionSelector);

            collectionItemsWrapper.insertAdjacentHTML('beforeend', newItemHtml);

            // Execute JS scripts embedded in prototype
            const collectionItems = collectionItemsWrapper.querySelectorAll('.field-collection-item');
            const lastElement = collectionItems[collectionItems.length - 1];
            lastElement.querySelectorAll('script').forEach(script => eval(script.innerHTML));

            // for complex collections of items, show the newly added item as not collapsed
            if (!isArrayCollection) {
                EaCollectionProperty.updateCollectionItemCssClasses(collection);
                const lastElementCollapseButton = lastElement.querySelector('.accordion-button');
                lastElementCollapseButton.classList.remove('collapsed');
                const lastElementBody = lastElement.querySelector('.accordion-collapse');
                lastElementBody.classList.add('show');
            }

            document.dispatchEvent(new Event('ea.collection.item-added'));
        });

        collection.classList.add('processed');
    },

    updateCollectionItemCssClasses: (collection) => {
        if (null === collection) {
            return;
        }

        const collectionItems = collection.querySelectorAll('.field-collection-item');
        collectionItems.forEach((item) => item.classList.remove('field-collection-item-first', 'field-collection-item-last'));

        const firstElement = collectionItems[0];
        if (undefined === firstElement) {
            return;
        }
        firstElement.classList.add('field-collection-item-first');

        const lastElement = collectionItems[collectionItems.length - 1];
        if (undefined === lastElement) {
            return;
        }
        lastElement.classList.add('field-collection-item-last');
    }
};
/* ------------------------- end field-collection.js ------------------------ */

/* -------------------------------------------------------------------------- */
/*                                   APP.JS                                   */
/* -------------------------------------------------------------------------- */
class App {
    #sidebarWidthLocalStorageKey;
    #contentWidthLocalStorageKey;

    constructor() {
        this.#sidebarWidthLocalStorageKey = 'ea/sidebar/width';
        this.#contentWidthLocalStorageKey = 'ea/content/width';

        this.#removeHashFormUrl();
        this.#createMainMenu();
        this.#createLayoutResizeControls();
        this.#createNavigationToggler();
        this.#createSearchHighlight();
        this.#createFilters();
        this.#createAutoCompleteFields();
        this.#createBatchActions();
        this.#createModalWindowsForDeleteActions();
        this.#createPopovers();
        // this.#createTooltips();

        document.addEventListener('ea.collection.item-added', () => this.#createAutoCompleteFields());
    }

    // When using tabs in forms, the selected tab is persisted (in the URL hash) so you
    // can see the same tab when reloading the page (e.g. '#tab-contact-information').
    // This method removes the hash from URL in the index page to not show form-related
    // information in the index page
    #removeHashFormUrl() {
        if (!window.location.href.includes('#')) {
            return;
        }

        // remove the hash only in the index page
        if (!document.querySelector('body').classList.contains('ea-index')) {
            return;
        }

        // don't set the hash to '' because that also removes the query parameters
        const urlParts = window.location.href.split('#');
        const urlWithoutHash = urlParts[0];
        window.history.replaceState({}, '', urlWithoutHash);
    }

    #createMainMenu() {
        // inspired by https://codepen.io/phileflanagan/pen/mwpQpY
        const menuItemsWithSubmenus = document.querySelectorAll('#main-menu .menu-item.has-submenu');
        menuItemsWithSubmenus.forEach((menuItem) => {
            const menuItemSubmenu = menuItem.querySelector('.submenu');
            if (null === menuItemSubmenu) {
                return;
            }

            // needed because the menu accordion is based on the max-height property.
            // visible elements must be initialized with a explicit max-height; otherwise
            // when you click on them the first time, the animation is not smooth
            if (menuItem.classList.contains('expanded')) {
                menuItemSubmenu.style.maxHeight = menuItemSubmenu.scrollHeight + 'px';
            }

            menuItem.querySelector('.submenu-toggle').addEventListener('click', (event) =>  {
                event.preventDefault();

                // hide other submenus
                menuItemsWithSubmenus.forEach((otherMenuItem) => {
                    if (menuItem === otherMenuItem) {
                        return;
                    }

                    const otherMenuItemSubmenu = otherMenuItem.querySelector('.submenu');
                    if (otherMenuItem.classList.contains('expanded')) {
                        otherMenuItemSubmenu.style.maxHeight = '0px';
                        otherMenuItem.classList.remove('expanded');
                    }
                });

                // toggle the state of this submenu
                if (menuItem.classList.contains('expanded')) {
                    menuItemSubmenu.style.maxHeight = '0px';
                    menuItem.classList.remove('expanded');
                } else {
                    menuItemSubmenu.style.maxHeight = menuItemSubmenu.scrollHeight + 'px';
                    menuItem.classList.add('expanded');
                }
            });
        });
    }

    #createLayoutResizeControls() {
        const sidebarResizerHandler = document.querySelector('#sidebar-resizer-handler');
        if (null !== sidebarResizerHandler) {
            sidebarResizerHandler.addEventListener('click', () => {
                const oldValue = localStorage.getItem(this.#sidebarWidthLocalStorageKey) || 'normal';
                const newValue = 'normal' === oldValue ? 'compact' : 'normal';

                document.querySelector('body').classList.remove(`ea-sidebar-width-${ oldValue }`);
                document.querySelector('body').classList.add(`ea-sidebar-width-${ newValue }`);
                localStorage.setItem(this.#sidebarWidthLocalStorageKey, newValue);
            });
        }

        const contentResizerHandler = document.querySelector('#content-resizer-handler');
        if (null !== contentResizerHandler) {
            contentResizerHandler.addEventListener('click', () => {
                const oldValue = localStorage.getItem(this.#contentWidthLocalStorageKey) || 'normal';
                const newValue = 'normal' === oldValue ? 'full' : 'normal';

                document.querySelector('body').classList.remove(`ea-content-width-${ oldValue }`);
                document.querySelector('body').classList.add(`ea-content-width-${ newValue }`);
                localStorage.setItem(this.#contentWidthLocalStorageKey, newValue);
            });
        }
    }

    #createNavigationToggler() {
        const toggler = document.querySelector('#navigation-toggler');
        const cssClassName = 'ea-mobile-sidebar-visible';
        let modalBackdrop;

        if (null === toggler) {
            return;
        }

        toggler.addEventListener('click', () => {
            document.querySelector('body').classList.toggle(cssClassName);

            if (document.querySelector('body').classList.contains(cssClassName)) {
                modalBackdrop = document.createElement('div');
                modalBackdrop.classList.add('modal-backdrop', 'fade', 'show');
                modalBackdrop.onclick = () => {
                    document.querySelector('body').classList.remove(cssClassName);
                    document.body.removeChild(modalBackdrop);
                    modalBackdrop = null;
                };

                document.body.appendChild(modalBackdrop);
            } else if (modalBackdrop) {
                document.body.removeChild(modalBackdrop);
                modalBackdrop = null;
            }
        });
    }

    #createSearchHighlight() {
        const searchElement = document.querySelector('.form-action-search [name="query"]');
        if (null === searchElement) {
            return;
        }

        const searchQuery = searchElement.value;
        if ('' === searchQuery.trim()) {
            return;
        }

        // splits a string into tokens, taking into account quoted strings
        // Example: 'foo "bar baz" qux' => ['foo', 'bar baz', 'qux']
        const tokenizeString = (string) => {
            const regex = /"([^"\\]*(\\.[^"\\]*)*)"|\S+/g;
            const tokens = [];
            let match;

            while (null !== (match = regex.exec(string))) {
                tokens.push(match[0].replaceAll('"', '').trim());
            }

            return tokens;
        };

        const searchQueryTerms = tokenizeString(searchElement.value);
        const searchQueryTermsHighlightRegexp = new RegExp(searchQueryTerms.join('|'), 'i');

        const elementsToHighlight = document.querySelectorAll('table tbody td.searchable');
        const highlighter = new markJs(elementsToHighlight);
        highlighter.markRegExp(searchQueryTermsHighlightRegexp);
    }

    #createFilters() {
        const filterButton = document.querySelector('.datagrid-filters .action-filters-button');
        if (null === filterButton) {
            return;
        }

        const filterModal = document.querySelector(filterButton.getAttribute('data-bs-target'));

        // this is needed to avoid errors when connection is slow
        filterButton.setAttribute('href', filterButton.getAttribute('data-href'));
        filterButton.removeAttribute('data-href');
        filterButton.classList.remove('disabled');

        filterButton.addEventListener('click', (event) => {
            const filterModalBody = filterModal.querySelector('.modal-body');
            filterModalBody.innerHTML = '<div class="fa-3x px-3 py-3 text-muted text-center"><i class="fas fa-circle-notch fa-spin"></i></div>';

            fetch(filterButton.getAttribute('href'))
                .then((response) => { return response.text(); })
                .then((text) => {
                    filterModalBody.innerHTML = text;
                    this.#createAutoCompleteFields();
                    this.#createFilterToggles();
                })
                .catch((error) => { console.error(error); });

            event.preventDefault();
        });

        const removeFilter = (filterField) => {
            filterField.closest('form').querySelectorAll(`input[name^="filters[${filterField.dataset.filterProperty}]"]`).forEach((filterFieldInput) => {
                filterFieldInput.remove();
            });

            filterField.remove();
        };

        document.querySelector('#modal-clear-button').addEventListener('click', () => {
            filterModal.querySelectorAll('.filter-field').forEach((filterField) => {
                removeFilter(filterField);
            });
            filterModal.querySelector('form').submit();
        });

        document.querySelector('#modal-apply-button').addEventListener('click', () => {
            filterModal.querySelectorAll('.filter-checkbox:not(:checked)').forEach((notAppliedFilter) => {
                removeFilter(notAppliedFilter.closest('.filter-field'));
            });
            filterModal.querySelector('form').submit();
        });
    }

    #createBatchActions() {
        console.log("Starting script execution...");
        const selectAllCheckbox = document.querySelector('.form-batch-checkbox-all');
        if (null === selectAllCheckbox) {
            console.log("No select all checkbox found, exiting...");
            return;
        } else {
            console.log("Select all checkbox found, continuing...");
        }

        console.log("Found form-batch-checkbox-all element:", selectAllCheckbox);
        const rowCheckboxes = document.querySelectorAll('input[type="checkbox"].form-batch-checkbox');
        selectAllCheckbox.addEventListener("change", function() {
            console.log("Change event listener attached to form-batch-checkbox-all element");
            rowCheckboxes.forEach(function(rowCheckbox) {
                console.log("Processing rowCheckbox:", rowCheckbox);
                rowCheckbox.checked = selectAllCheckbox.checked;
                console.log("Set rowCheckbox.checked to:", selectAllCheckbox.checked);
                rowCheckbox.dispatchEvent(new Event("change"));
                console.log("Dispatched 'change' event for checkbox");
            });
        });

        var deselectAllButton = document.querySelector(".deselect-batch-button");
        if (null !== deselectAllButton) {
            console.log("Found deselect-batch-button element:", deselectAllButton);
            deselectAllButton.addEventListener("click", function() {
                console.log("Click event listener attached to deselect-batch-button element");
                selectAllCheckbox.checked = false;
                console.log("Set form-batch-checkbox-all checked state to false");
                selectAllCheckbox.dispatchEvent(new Event("change"));
                console.log("Dispatched 'change' event for form-batch-checkbox-all element");
            });
        }

        rowCheckboxes.forEach(function(checkbox, index) {
            console.log("Processing checkbox at index:", index);
            
            checkbox.addEventListener("change", function() {
                console.log("Change event listener attached to checkbox at index:", index);
                console.log("Row checkbox changed, updating UI...");
                const selectedRowCheckboxes = document.querySelectorAll('input[type="checkbox"].form-batch-checkbox:checked');
                const row = checkbox.closest('tr');
                const content = checkbox.closest('.nk-content');
                console.log('Content: ', content);

                if (checkbox.checked) {
                    row.classList.add("selected-row");
                    console.log("Added 'selected-row' class to tr element");
                } else {
                    row.classList.remove("selected-row");
                    selectAllCheckbox.checked = false;
                    console.log("Removed 'selected-row' class from tr element and set form-batch-checkbox-all checked state to false");
                }

                const rowsAreSelected = 0 !== selectedRowCheckboxes.length;
                const contentTitle = document.querySelector('.content-header-title > .title');

                if (content !== null) {
                    const globalActions = content.querySelector('.global-actions');
                    const batchActions = content.querySelector('.batch-actions');

                    if (null !== globalActions) {
                        toggleVisibilityClasses(globalActions, rowsAreSelected);
                    }
                    if (null !== batchActions) {
                        toggleVisibilityClasses(batchActions, !rowsAreSelected);
                    }
                }                
        
                if (null !== contentTitle) {
                    toggleVisibilityClasses(contentTitle, rowsAreSelected);
                }
            });
        });

        const modalTitle = document.querySelector('#batch-action-confirmation-title');
        const titleContentWithPlaceholders = modalTitle.textContent;
        
        document.querySelectorAll('[data-action-batch]').forEach((dataActionBatch) => {
            dataActionBatch.addEventListener('click', (event) => {
                console.log("Batch action button clicked, preparing batch action...");
                event.preventDefault();
        
                const actionElement = event.currentTarget;
                const actionName = actionElement.textContent.trim() || actionElement.getAttribute('title');
                const selectedItems = document.querySelectorAll('input[type="checkbox"].form-batch-checkbox:checked');
                modalTitle.textContent = titleContentWithPlaceholders
                    .replace('%action_name%', actionName)
                    .replace('%num_items%', selectedItems.length.toString());
        
                document.querySelector('#modal-batch-action-button').addEventListener('click', () => {
                    console.log("Modal batch action button clicked, submitting form...");
                    // prevent double submission of the batch action form
                    actionElement.setAttribute('disabled', 'disabled');
        
                    const batchFormFields = {
                        'batchActionName': actionElement.getAttribute('data-action-name'),
                        'entityFqcn': actionElement.getAttribute('data-entity-fqcn'),
                        'batchActionUrl': actionElement.getAttribute('data-action-url'),
                        'batchActionCsrfToken': actionElement.getAttribute('data-action-csrf-token'),
                    };
                    selectedItems.forEach((item, i) => {
                        batchFormFields[`batchActionEntityIds[${i}]`] = item.value;
                    });
        
                    const batchForm = document.createElement('form');
                    batchForm.setAttribute('method', 'POST');
                    batchForm.setAttribute('action', actionElement.getAttribute('data-action-url'));
                    for (let fieldName in batchFormFields) {
                        const formField = document.createElement('input');
                        formField.setAttribute('type', 'hidden');
                        formField.setAttribute('name', fieldName);
                        formField.setAttribute('value', batchFormFields[fieldName]);
                        batchForm.appendChild(formField);
                    }
        
                    document.body.appendChild(batchForm);
                    batchForm.submit();
                });
            });
        });
    }

    #createAutoCompleteFields() {
        const autocomplete = new Autocomplete();
        document.querySelectorAll('[data-ea-widget="ea-autocomplete"]').forEach((autocompleteElement) => {
            autocomplete.create(autocompleteElement);
        });
    }

    #createModalWindowsForDeleteActions() {
        document.querySelectorAll('.action-delete').forEach((actionElement) => {
            actionElement.addEventListener('click', (event) => {
                event.preventDefault();

                document.querySelector('#modal-delete-button').addEventListener('click', () => {
                    const deleteFormAction = actionElement.getAttribute('formaction');
                    const deleteForm = document.querySelector('#delete-form');
                    deleteForm.setAttribute('action', deleteFormAction);
                    deleteForm.submit();
                });
            });
        });
    }

    #createPopovers() {
        document.querySelectorAll('[data-bs-toggle="popover"]').forEach((popoverElement) => {
            new bootstrap.Popover(popoverElement);
        });
    }

    #createTooltips() {
        document.querySelectorAll('[data-bs-toggle="tooltip"]').forEach((tooltipElement) => {
            new bootstrap.Tooltip(tooltipElement);
        });
    }

    #createFilterToggles() {
        document.querySelectorAll('.filter-checkbox').forEach((filterCheckbox) => {
            filterCheckbox.addEventListener('change', () => {
                const filterToggleLink = filterCheckbox.nextElementSibling;
                const filterExpandedAttribute = filterCheckbox.nextElementSibling.getAttribute('aria-expanded');

                if ((filterCheckbox.checked && 'false' === filterExpandedAttribute) || (!filterCheckbox.checked && 'true' === filterExpandedAttribute)) {
                    filterToggleLink.click();
                }
            });
        });

        document.querySelectorAll('form[data-ea-filters-form-id]').forEach((form) => {
            // TODO: when using the native datepicker, 'change' isn't fired unless you input the entire date + time information
            form.addEventListener('change', (event) => {
                if (event.target.classList.contains('filter-checkbox')) {
                    return;
                }

                const filterCheckbox = event.target.closest('.filter-field').querySelector('.filter-checkbox');
                if (!filterCheckbox.checked) {
                    filterCheckbox.checked = true;
                }
            });
        });

        document.querySelectorAll('[data-ea-comparison-id]').forEach((comparisonWidget) => {
            comparisonWidget.addEventListener('change', (event) => {
                const comparisonWidget = event.currentTarget;
                const comparisonId = comparisonWidget.dataset.eaComparisonId;

                if (comparisonId === undefined) {
                    return;
                }

                const secondValue = document.querySelector(`[data-ea-value2-of-comparison-id="${comparisonId}"]`);

                if (secondValue === null) {
                    return;
                }

                toggleVisibilityClasses(secondValue, comparisonWidget.value !== 'between');
            });
        });
    }
}
/* ------------------------------- end app.js ------------------------------- */

/* -------------------------------------------------------------------------- */
/*                               AUTOCOMPLETE.JS                              */
/* -------------------------------------------------------------------------- */
class Autocomplete
{
    create(element) {
        // this avoids initializing the same field twice (TomSelect shows an error otherwise)
        if (element.classList.contains('tomselected')) {
            return;
        }

        const autocompleteEndpointUrl = element.getAttribute('data-ea-autocomplete-endpoint-url');
        if (null !== autocompleteEndpointUrl) {
            return this.#createAutocompleteWithRemoteData(element, autocompleteEndpointUrl);
        }

        const renderOptionsAsHtml = 'true' === element.getAttribute('data-ea-autocomplete-render-items-as-html');
        if (renderOptionsAsHtml) {
            return this.#createAutocompleteWithHtmlContents(element);
        }

        return this.#createAutocomplete(element);
    }

    #getCommonConfig(element) {
        const config = {
            render: {
                no_results: function(data, escape) {
                    return `<div class="no-results">${element.getAttribute('data-ea-i18n-no-results-found')}</div>`;
                },
            },
            plugins: {
                dropdown_input: {},
            }
        };

        if (null === element.getAttribute('required') && null === element.getAttribute('disabled')) {
            config.plugins.clear_button = { title: '' };
        }

        if (null !== element.getAttribute('multiple')) {
            config.plugins.remove_button = { title: '' };
        }

        if (null !== element.getAttribute('data-ea-autocomplete-endpoint-url')) {
            config.plugins.virtual_scroll = {};
        }

        if ('true' === element.getAttribute('data-ea-autocomplete-allow-item-create')) {
            config.create = true;
        }

        return config;
    };

    #createAutocomplete(element) {
        const config = this.#mergeObjects(this.#getCommonConfig(element), {
            maxOptions: null,
        });

        return new TomSelect(element, config);
    }

    #createAutocompleteWithHtmlContents(element) {
        const autoSelectOptions = [];
        for (let i = 0; i < element.options.length; i++) {
            const label = element.options[i].text;
            const value = element.options[i].value;

            autoSelectOptions.push({
                label_text: this.#stripTags(label),
                label_raw: label,
                value: value,
            });
        }

        const config = this.#mergeObjects(this.#getCommonConfig(element), {
            valueField: 'value',
            labelField: 'label_raw',
            searchField: ['label_text'],
            options: autoSelectOptions,
            maxOptions: null,
            render: {
                item: function(item, escape) {
                    return `<div>${item.label_raw}</div>`;
                },
                option: function(item, escape) {
                    return `<div>${item.label_raw}</div>`;
                }
            },
        });

        return new TomSelect(element, config);
    }

    #createAutocompleteWithRemoteData(element, autocompleteEndpointUrl) {
        const config = this.#mergeObjects(this.#getCommonConfig(element), {
            valueField: 'entityId',
            labelField: 'entityAsString',
            searchField: ['entityAsString'],
            firstUrl: (query) => {
                return autocompleteEndpointUrl + '&query=' + encodeURIComponent(query);
            },
            // VERY IMPORTANT: use 'function (query, callback) { ... }' instead of the
            // '(query, callback) => { ... }' syntax because, otherwise,
            // the 'this.XXX' calls inside of this method fail
            load: function (query, callback) {
                const url = this.getUrl(query);
                fetch(url)
                    .then(response => response.json())
                    // important: next_url must be set before invoking callback()
                    .then(json => { this.setNextUrl(query, json.next_page); callback(json.results) })
                    .catch(() => callback());
            },
            preload: 'focus',
            maxOptions: null,
            // on remote calls, we don't want tomselect to further filter the results by "entityAsString"
            // this override causes all results to be returned with the sorting from the server
            score: function(search) {
                return function(item) {
                    return 1;
                };
            },
            render: {
                option: function(item, escape) {
                    return `<div>${escape(item.entityAsString)}</div>`;
                },
                item: function(item, escape) {
                    return `<div>${escape(item.entityAsString)}</div>`;
                },
                loading_more: function(data, escape) {
                    return `<div class="loading-more-results">${element.getAttribute('data-ea-i18n-loading-more-results')}</div>`;
                },
                no_more_results: function(data, escape) {
                    return `<div class="no-more-results">${element.getAttribute('data-ea-i18n-no-more-results')}</div>`;
                },
                no_results: function(data, escape) {
                    return `<div class="no-results">${element.getAttribute('data-ea-i18n-no-results-found')}</div>`;
                },
            },
        });

        return new TomSelect(element, config);
    }

    #stripTags(string) {
        return string.replace(/(<([^>]+)>)/gi, '');
    }

    #mergeObjects(object1, object2) {
        return { ...object1, ...object2 };
    }
}
/* --------------------------- end autocomplete.js -------------------------- */

/* -------------------------------------------------------------------------- */
/*                              FIELD-BOOLEAN.JS                              */
/* -------------------------------------------------------------------------- */
class ToggleSwitch {
    constructor(field) {
        this.field = field;
        this.field.addEventListener('change', this.#updateFieldValue.bind(this));
    }

    #updateFieldValue() {
        const newValue = this.field.checked;
        const toggleUrl = this.field.getAttribute('data-toggle-url') + "&newValue=" + newValue.toString();

        fetch(toggleUrl, {
            method: 'PATCH',
            // the XMLHttpRequest header is needed to keep compatibility with the previous code, which didn't use the Fetch API
            headers: { 'X-Requested-With': 'XMLHttpRequest' },
        })
        .then((response) => {
            if (!response.ok) {
                this.#disableField();
            }

            return response.text();
        })
        .then(() => { /* do nothing else when the toggle request is successful */ })
        .catch(() => this.#disableField());
    }

    // used in case of error, to restore the original toggle field value and disable it
    #disableField() {
        this.field.checked = !this.field.checked;
        this.field.disabled = true;
        this.field.closest('.form-switch').classList.add('disabled');
    }
}
/* -------------------------- end field-boolean.js -------------------------- */

/* -------------------------------------------------------------------------- */
/*                            FIELD-FILE-UPLOAD.JS                            */
/* -------------------------------------------------------------------------- */
const eaFileUploadHandler = function (event) {
    document.querySelectorAll('.ea-fileupload input[type="file"]').forEach((fileUploadElement) => {
        new FileUploadField(fileUploadElement);
    });
}

window.addEventListener('DOMContentLoaded', eaFileUploadHandler);
document.addEventListener('ea.collection.item-added', eaFileUploadHandler);

class FileUploadField {
    #fieldContainerElement;

    constructor(field) {
        this.field = field;
        this.#fieldContainerElement = this.field.closest('.ea-fileupload');
        this.field.addEventListener('change', this.#updateField.bind(this));

        let deleteButton = this.#getFieldDeleteButton();
        if (deleteButton) {
            deleteButton.addEventListener('click', this.#resetField.bind(this));
        }
    }

    #updateField() {
        if (0 === this.field.files.length) {
            return;
        }

        const filename = (1 === this.field.files.length) ? this.field.files[0].name : this.field.files.length + ' ' + this.field.getAttribute('data-files-label');

        let totalSizeInBytes = 0;
        for (const file of this.field.files) {
            totalSizeInBytes += file.size;
        }

        this.#getFieldCustomInput().innerHTML = filename;
        this.#getFieldDeleteButton().style.display = 'block';
        this.#getFieldSizeLabel().childNodes.forEach((fileUploadFileSizeLabelChild) => {
            if (fileUploadFileSizeLabelChild.nodeType === Node.TEXT_NODE) {
                this.#getFieldSizeLabel().removeChild(fileUploadFileSizeLabelChild);
            }
        });
        this.#getFieldSizeLabel().prepend(this.#humanizeFileSize(totalSizeInBytes));
    }

    #resetField() {
        const fieldDeleteCheckbox = this.#fieldContainerElement.querySelector('input[type=checkbox].form-check-input');
        const fieldListOfFiles = this.#fieldContainerElement.querySelector('.fileupload-list');

        if (fieldDeleteCheckbox) {
            fieldDeleteCheckbox.checked = true;
            fieldDeleteCheckbox.click();
        }
        this.field.value = '';
        this.#getFieldCustomInput().innerHTML = '';
        toggleVisibilityClasses(this.#getFieldDeleteButton(), true);

        this.#getFieldSizeLabel().childNodes.forEach((fileSizeLabelChild) => {
            if (fileSizeLabelChild.nodeType === Node.TEXT_NODE) {
                this.#getFieldSizeLabel().removeChild(fileSizeLabelChild);
            }
        });

        if (null !== fieldListOfFiles) {
            toggleVisibilityClasses(fieldListOfFiles, true);
        }
    }

    #humanizeFileSize(bytes) {
        const unit = ['B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'];
        const factor = Math.trunc(Math.floor(Math.log(bytes) / Math.log(1024)));

        return Math.trunc(bytes / (1024 ** factor)) + unit[factor];
    }

    #getFieldCustomInput() {
        return this.#fieldContainerElement.querySelector('.custom-file-label');
    }

    #getFieldDeleteButton() {
        return this.#fieldContainerElement.querySelector('.ea-fileupload-delete-btn');
    }

    #getFieldSizeLabel() {
        return this.#fieldContainerElement.querySelector('.input-group-text');
    }
}
/* ------------------------ end field-file-upload.js ------------------------ */

/* -------------------------------------------------------------------------- */
/*                               FIELD-IMAGE.JS                               */
/* -------------------------------------------------------------------------- */
class Image {
    constructor(field) {
        this.field = field;
        this.field.addEventListener('click', this.#renderLightbox.bind(this));
    }

    #renderLightbox() {
        console.log('renderLightbox called');
        const lightboxContent = document.querySelector(this.field.getAttribute('data-ea-lightbox-content-selector')).innerHTML;
        const lightbox = basicLightbox.create(lightboxContent);
        lightbox.show();
    }
}
/* --------------------------- end field-image.js --------------------------- */

/* -------------------------------------------------------------------------- */
/*                                FIELD-SLUG.JS                               */
/* -------------------------------------------------------------------------- */
slugify.extend({
    "$": "",
    "%": "",
    "&": "",
    "<": "",
    ">": "",
    "|": "",
    "¢": "",
    "£": "",
    "¤": "",
    "¥": "",
    "₠": "",
    "₢": "",
    "₣": "",
    "₤": "",
    "₥": "",
    "₦": "",
    "₧": "",
    "₨": "",
    "₩": "",
    "₪": "",
    "₫": "",
    "€": "",
    "₭": "",
    "₮": "",
    "₯": "",
    "₰": "",
    "₱": "",
    "₲": "",
    "₳": "",
    "₴": "",
    "₵": "",
    "₸": "",
    "₹": "",
    "₽": "",
    "₿": "",
    "∂": "",
    "∆": "",
    "∑": "",
    "∞": "",
    "♥": "",
    "元": "",
    "円": "",
    "﷼": "",
});

class Slugger {
    constructor(field) {
        this.field = field;
        this.setTargetElement();
        this.locked = true;
        this.field.setAttribute('readonly', 'readonly');

        if ('' === this.field.value) {
            this.currentSlug = '';
            this.updateValue();
            this.listenTarget();
        } else {
            this.currentSlug = this.field.value;
        }

        this.appendLockButton();
    }

    setTargetElement() {
        const fieldNames = JSON.parse(this.field.dataset.target);
        this.targets = [];

        for (const name of fieldNames) {
            const target = document.getElementById(name);

            if (null === target) {
                throw `Wrong target specified for slug widget ("${name}").`;
            }

            this.targets.push(target);
        }
    }

    /**
     * Append a "lock" button to control slug behaviour (auto or manual)
     */
    appendLockButton() {
        this.lockButton = this.field.parentNode.querySelector('button');
        this.lockButtonIcon = this.lockButton.querySelector('i');
        this.lockButton.addEventListener('click', () => {
            if (this.locked) {
                let confirmMessage = this.field.dataset.confirmText || null;
                if (null === confirmMessage) {
                    this.unlock();
                } else {
                    let formattedConfirmMessage = decodeURIComponent(JSON.parse('"' + confirmMessage.replace(/\"/g, '\\"') + '"'));
                    if (true === confirm(formattedConfirmMessage)) {
                        this.unlock();
                    }
                }
            } else {
                this.lock();
            }
        });
    }

    /**
     * Unlock the widget input (manual mode)
     */
    unlock() {
        this.locked = false;
        this.lockButtonIcon.classList.replace('fa-lock', 'fa-lock-open');
        this.field.removeAttribute('readonly');
    }

    /**
     * Lock the widget input (auto mode)
     */
    lock() {
        this.locked = true;
        this.lockButtonIcon.classList.replace('fa-lock-open', 'fa-lock');

        // Locking it back changes the value either to default value, or recomputes it
        if ('' !== this.currentSlug) {
            this.field.value = this.currentSlug;
        } else {
            this.updateValue();
        }

        this.field.setAttribute('readonly', 'readonly');
    }

    updateValue() {
        this.field.value = slugify(this.targets.map(target => target.value).join('-'), {
            remove: /[^A-Za-z0-9\s-]/g,
            lower: true,
            strict: true,
        });
    }

    /**
     * Observe the target field and slug it
     */
    listenTarget() {
        for (const target of this.targets) {
            target.addEventListener('keyup', () => {
                if ('readonly' === this.field.getAttribute('readonly')) {
                    this.updateValue();
                }
            });
        }
    }
}
/* ---------------------------- end field-slug.js --------------------------- */

/* -------------------------------------------------------------------------- */
/*                            FIELD-TEXT-EDITOR.JS                            */
/* -------------------------------------------------------------------------- */
// Listening to the DOMLoadedContent event is too late because the Trix editor is already initialized.
// To be sure to handle properly the custom configuration, we have to to listen to the trix-before-initialize event.
document.addEventListener('trix-before-initialize', () => {
    new TextEditorField();
});

class TextEditorField {
    constructor() {
        this.#processRequiredAttribute();
        this.#enableFormChangesDetection();
        this.#handleFormSubmission();
    }

    #processRequiredAttribute() {
        // TrixEditor works by storing the original content in a hidden <textarea> and creating a new <trix-editor> element
        // When the original field is required and its content is empty, browsers try to add a validation error, but it
        // fails because the element is hidden ("An invalid form control with name=”foo" is not focusable")
        // That's why we remove the HTML required attribute and store it in a custom attribute that will be check later
        document.querySelectorAll('textarea.ea-text-editor-content').forEach((trixContentElement) => {
            const trixEditorConfig = JSON.parse(trixContentElement.getAttribute('data-trix-editor-config'));
            if (null !== trixEditorConfig) {
                Trix.config = this.#mergeObjects(trixEditorConfig, Trix.config);
            }

            const isTrixFieldRequired = 'required' === trixContentElement.getAttribute('required') ? 'true' : 'false';
            trixContentElement.setAttribute('data-ea-trix-is-required', isTrixFieldRequired);
            trixContentElement.removeAttribute('required');

            // Change number of rows
            if (trixContentElement.dataset.numberOfRows !== '') {
                const editor = document.querySelector(`trix-editor[input=${trixContentElement.id}].trix-content`);

                if (editor !== null) {
                    // Here we consider 21px as the average line height
                    editor.style.setProperty('min-height', `${21 * trixContentElement.dataset.numberOfRows}px`);
                }
            }
        });

        // Since the above code is needed to remove the HTML required attribute, we need to add a custom validity message
        // for the required fields because otherwise the validation method will not add an error badge to the related tabs
        // and the submit button will be disabled and never be enabled again.
        this.#markInvalidFormFields();
        document.addEventListener('trix-change', () => {
            this.#markInvalidFormFields();
        });
    }

    #enableFormChangesDetection() {
        // Because of the way TrixEditor works, browsers cannot detect changes to these fields automatically,
        // so we manually trigger the DirtyForm plugin when the content changes.
        document.addEventListener('trix-change', function (event) {
            const form = event.target.closest('form');
            if (null === form) {
                return;
            }

            new DirtyForm(form);
        });
    }

    #handleFormSubmission() {
        document.addEventListener('ea.form.submit', (formEvent) => {
            const entityForm = formEvent.detail.form;
            entityForm.querySelectorAll('textarea.ea-text-editor-content').forEach(function (trixContentElement) {
                const isTrixFieldRequired = 'true' === trixContentElement.getAttribute('data-ea-trix-is-required');
                const trixEditorElement = entityForm.querySelector(`trix-editor[input=${trixContentElement.id}]`);
                // an empty Trix editor field is not really empty; it contains a "\n" character (%0A = HTML encoded)
                const isTrixEditorEmpty = '%0A' === escape(trixEditorElement.editor.getDocument().toString());

                if (isTrixFieldRequired && isTrixEditorEmpty) {
                    const formGroup = trixContentElement.closest('div.form-group');
                    formGroup.classList.add('has-error');
                    formGroup.addEventListener('click', function onFormGroupClick() {
                        formGroup.classList.remove('has-error');
                        formGroup.removeEventListener('click', onFormGroupClick);
                    });

                    const errorMessage = this.#getLocalizedErrorMessage();
                    let errorElement = document.createElement('div');
                    errorElement.classList.add('invalid-feedback', 'd-block');
                    errorElement.innerHTML = `<span class="form-error-message">${ errorMessage }</span>`;
                    trixContentElement.closest('.form-widget').append(errorElement);

                    formEvent.preventDefault();
                }
            });
        });
    }

    #markInvalidFormFields() {
        ['.ea-new-form', '.ea-edit-form'].forEach((formSelector) => {
            const form = document.querySelector(formSelector);
            if (null !== form) {
                form.querySelectorAll('input,select,textarea').forEach((input) => {
                    if (
                        input.hasAttribute('data-ea-trix-is-required') &&
                        input.getAttribute('data-ea-trix-is-required') === 'true' &&
                        input.value === ''
                    ) {
                        input.setCustomValidity('invalid');
                    } else {
                        input.setCustomValidity('');
                    }
                });
            }
        });
    };

    #getLocalizedErrorMessage() {
        // copied from https://github.com/chromium/chromium/search?p=1&q=2507943997699731163
        const requiredFieldMessage = {
            'ar': 'يُرجى ملء هذا الحقل.',
            'bg': 'Моля, попълнете това поле',
            'ca': 'Empleneu aquest camp',
            'cs': 'Vyplňte prosím toto pole',
            'da': 'Udfyld dette felt',
            'de': 'Füllen Sie dieses Feld aus',
            'el': 'Συμπληρώστε αυτό το πεδίο',
            'en': 'Please fill in this field',
            'es': 'Completa este campo',
            'eu': 'Bete eremu hau',
            'fa': 'لطفاً این قسمت را تکمیل کنید.',
            'fi': 'Täytä tämä kenttä',
            'fr': 'Veuillez renseigner ce champ',
            'gl': 'Completa este campo',
            'hr': 'Ispunite ovo polje',
            'hu': 'Kérjük, töltse ki ezt a mezőt',
            'it': 'Compila questo campo',
            'lt': 'Užpildykite šį lauką',
            'nl': 'Vul dit veld in',
            'no': 'Vennligst fyll ut dette feltet',
            'pl': 'Wypełnij to pole',
            'pt': 'Preencha este campo',
            'pt_BR': 'Preencha este campo',
            'ro': 'Completează acest câmp',
            'ru': 'Заполните это поле',
            'sl': 'Izpolnite to polje',
            'sr_RS': 'Попуните ово поље',
            'sv': 'Fyll i det här fältet',
            'tr': 'Lütfen bu alanı doldurun',
            'uk': 'Заповніть це поле',
            'zh_CN': '请填写此字段',
        };

        return requiredFieldMessage[document.querySelector('html').getAttribute('lang')] || 'Please fill in this field';
    }

    // copied from https://gist.github.com/ahtcx/0cd94e62691f539160b32ecda18af3d6?permalink_comment_id=3889214#gistcomment-3889214
    #mergeObjects(fromObject, intoObject) {
        for (const [key, val] of Object.entries(fromObject)) {
            if (val !== null && typeof val === `object`) {
                if (intoObject[key] === undefined) {
                    intoObject[key] = new val.__proto__.constructor();
                }
                this.#mergeObjects(val, intoObject[key]);
            } else {
                intoObject[key] = val;
            }
        }

        return intoObject;
    }
}
/* ------------------------ end field-text-editor.js ------------------------ */

/* -------------------------------------------------------------------------- */
/*                              FIELD-TEXTAREA.JS                             */
/* -------------------------------------------------------------------------- */
class Autogrow {
    constructor(field) {
        this.field = field;
        this.field.addEventListener('input', this.autogrow.bind(this));
        this.autogrow();
    }

    autogrow() {
        this.field.style.overflow = 'hidden';
        this.field.style.resize = 'none';
        this.field.style.boxSizing = 'border-box';
        this.field.style.height = 'auto';

        // this check is needed because the <textarea> element can be inside a
        // minimizable panel, causing its scrollHeight value to be 0
        if (this.field.scrollHeight > 0) {
            this.field.style.height = this.field.scrollHeight + 'px';
        }
    }
}
/* -------------------------- end field-textarea.js ------------------------- */

/* -------------------------------------------------------------------------- */
/*                                   FORM.JS                                  */
/* -------------------------------------------------------------------------- */
class Form {
    constructor() {
        this.#persistSelectedTab();
        this.#createUnsavedFormChangesWarning();
        this.#createFieldsWithErrors();
        this.#preventMultipleFormSubmission();
    }

    #persistSelectedTab() {
        // the ID of the selected tab is appended as a hash in the URL to persist it;
        // if the URL has a hash, try to look for a tab with that ID and show it
        const urlHash = window.location.hash;
        if (urlHash) {
            const selectedTabPaneId = urlHash.substring(1); // remove the leading '#' from the hash
            const selectedTabId = `tablist-${ selectedTabPaneId }`;
            this.#setTabAsActive(selectedTabId);
        }

        // update the page anchor when the selected tab changes
        document.querySelectorAll('a[data-bs-toggle="tab"]').forEach((tabElement) => {
            tabElement.addEventListener('shown.bs.tab', function (event) {
                const urlHash = '#' + event.target.getAttribute('href').substring(1);
                history.pushState({}, '', urlHash);
            });
        });
    }

    #createUnsavedFormChangesWarning() {
        ['.ea-new-form', '.ea-edit-form'].forEach((formSelector) => {
            const form = document.querySelector(formSelector);
            if (null === form) {
                return;
            }

            // although DirtyForm supports passing a custom message to display,
            // modern browsers don't allow to display custom messages to protect users
            new DirtyForm(form);
        });
    }

    #createFieldsWithErrors() {
        const handleFieldsWithErrors = (form, pageName) => {
            // Intercept errors before submit to avoid browser error "An invalid form control with name='...' is not focusable."
            //
            // Adding visual feedback for invalid fields: any ".form-group" with invalid fields
            // receives "has-error" class. The class is removed on click on the ".form-group"
            // itself to support custom/complex fields.
            //
            // Adding visual error counter feedback for invalid fields inside form tabs (visible or not)
            const that = this;
            document.querySelector('.ea-edit, .ea-new').querySelectorAll('[type="submit"]').forEach((button) => {
                button.addEventListener('click', function onSubmitButtonsClick(clickEvent) {
                    let formHasErrors = false;

                    // Remove all error counter badges
                    document.querySelectorAll('.form-tabs-tablist .nav-item .badge-danger.badge').forEach( (badge) => {
                        badge.parentElement.removeChild(badge);
                    });

                    if (null !== form.getAttribute('novalidate')) {
                        return;
                    }

                    form.querySelectorAll('input, select, textarea').forEach((input) => {
                        if (!input.disabled && !input.validity.valid) {
                            formHasErrors = true;

                            // Visual feedback for tabs
                            // Adding a badge with a error count next to the tab label
                            const formTab = input.closest('div.tab-pane');
                            if (formTab) {
                                // Match tab link either by "data-bs-target" attribute or by href linking to the id anchor
                                const navLinkTab = document.querySelector(`[data-bs-target="#${ formTab.id }"], a[href="#${ formTab.id }"]`);

                                if (navLinkTab) {
                                    navLinkTab.classList.add('has-error');

                                    const badge = navLinkTab.querySelector('.badge');
                                    if (badge) {
                                        // Increment number of error
                                        badge.textContent = (parseInt(badge.textContent) + 1).toString();
                                    } else {
                                        // Create a new badge
                                        let newErrorBadge = document.createElement('span');
                                        newErrorBadge.classList.add('badge', 'badge-danger');
                                        newErrorBadge.textContent = '1';
                                        navLinkTab.appendChild(newErrorBadge);
                                    }
                                }
                            }

                            // Visual feedback for group
                            const formGroup = input.closest('div.form-group');
                            formGroup.classList.add('has-error');

                            formGroup.addEventListener('click', function onFormGroupClick() {
                                formGroup.classList.remove('has-error');
                                formGroup.removeEventListener('click', onFormGroupClick);
                            });
                        }
                    });

                    if (formHasErrors) {
                        clickEvent.preventDefault();
                        clickEvent.stopPropagation();

                        // set as active the first tab with errors
                        const firstTabWithErrors = document.querySelector('.form-tabs-tablist .nav-tabs .nav-item .nav-link.has-error');
                        if (null !== firstTabWithErrors) {
                            that.#setTabAsActive(firstTabWithErrors.id);
                        }

                        document.dispatchEvent(new CustomEvent('ea.form.error', {
                            cancelable: true,
                            detail: { page: pageName, form: form }
                        }));
                    }
                });
            });

            form.addEventListener('submit', (submitEvent) => {
                const eaEvent = new CustomEvent('ea.form.submit', {
                    cancelable: true,
                    detail: { page: pageName, form: form }
                });
                const eaEventResult = document.dispatchEvent(eaEvent);
                if (false === eaEventResult) {
                    submitEvent.preventDefault();
                    submitEvent.stopPropagation();
                }
            });
        };

        ['.ea-new-form', '.ea-edit-form'].forEach((formSelector) => {
            const form = document.querySelector(formSelector);
            if (null !== form) {
                handleFieldsWithErrors(form, formSelector.includes('-new-') ? 'new' : 'edit');
            }
        });
    }

    #setTabAsActive(tabItemId) {
        const tabElement = document.getElementById(tabItemId);
        if (!tabElement) {
            return;
        }

        const Tab = bootstrap.Tab;
        const bootstrapTab = new Tab(tabElement);
        // when showing a tab, Bootstrap hides all the other tabs automatically
        bootstrapTab.show();
    }

    #preventMultipleFormSubmission() {
        ['.ea-new-form', '.ea-edit-form'].forEach((formSelector) => {
            const form = document.querySelector(formSelector);
            if (null === form) {
                return;
            }

            form.addEventListener('submit', () => {
                // this timeout is needed to include the disabled button into the submitted form
                setTimeout(() => {
                    const submitButtons = document.querySelector('.ea-edit, .ea-new').querySelectorAll('[type="submit"]');
                    submitButtons.forEach((button) => {
                        button.setAttribute('disabled', 'disabled');
                    });
                }, 1);
            }, false);
        });
    }
}
/* ------------------------------- end form.js ------------------------------ */

/* -------------------------------------------------------------------------- */
/*                                 HELPERS.JS                                 */
/* -------------------------------------------------------------------------- */
function toggleVisibilityClasses(element, removeVisibility) {
    if (removeVisibility) {
        element.classList.remove('d-block')
        element.classList.add('d-none')
    } else {
        element.classList.remove('d-none')
        element.classList.add('d-block')
    }
}
/* ----------------------------- end helpers.js ----------------------------- */
